import csv
import os
import time

from kehe_fl.utils.common.adafruit_scd_30 import AdafruitSCD30
from kehe_fl.utils.common.data_storage import DataStorage

class DataCollectionService:
    def __init__(self, fields, path: str, interval: int):
        self.sensor = AdafruitSCD30()
        self.storage = DataStorage(fields=fields, directory=path)
        self.path = path
        self.interval = interval
        self._running = False

    def start(self):
        print("[DataCollectionService] Starting data collection...")
        self._running = True
        while self._running:
            data = self.sensor.read()
            if data:
                co2, temperature, humidity = data
                self.storage.save({
                    "timestamp": time.strftime("%Y-%m-%d %H:%M:%S", time.localtime()),
                    "co2": co2,
                    "temperature": temperature,
                    "humidity": humidity
                })
            time.sleep(self.interval)

    def stop(self):
        print("[DataCollectionService] Stopping data collection...")
        self._running = False

    def check_data_count(self):
        print("[DataCollectionService] Checking data count...")
        files = os.listdir(self.path)
        count = 0
        for file in files:
            if file.endswith(".csv"):
                with open(os.path.join(self.path, file), 'r') as f:
                    reader = csv.reader(f)
                    count += sum(1 for _ in reader) - 1
        return count



import csv
import os
from datetime import datetime

class DataStorage:
    def __init__(self, fields, directory="logs"):
        self.fields = fields
        self.directory = os.path.expanduser(directory)
        os.makedirs(self.directory, exist_ok=True)

    def _get_daily_filepath(self):
        date_str = datetime.utcnow().strftime("%Y-%m-%d")
        return os.path.join(self.directory, f"{date_str}.csv")

    def save(self, data: dict):
        if not all(field in data for field in self.fields):
            raise ValueError("Missing fields in data: expected " + ", ".join(self.fields))

        filepath = self._get_daily_filepath()
        write_header = not os.path.isfile(filepath)

        with open(filepath, mode="a", newline="") as file:
            writer = csv.DictWriter(file, fieldnames=self.fields)
            if write_header:
                writer.writeheader()
            writer.writerow(data)



import json

import aiomqtt
import asyncio
from typing import Optional, Union

import numpy as np
from aiomqtt import Message

class MQTTProvider:
    def __init__(self, broker: str, port=1883, username: Optional[str] = None, password: Optional[str] = None):
        self.broker = broker
        self.port = port
        self.username = username
        self.password = password
        self._is_connected = False
        self._lock = asyncio.Lock()

    async def connect_and_listen(self):
        config = {"hostname": self.broker, "port": self.port}
        if self.username and self.password:
            config["username"] = self.username
            config["password"] = self.password

        async with aiomqtt.Client(**config) as self.client:
            print(f"[MQTT] Connected to {self.broker}:{self.port}")
            await self.subscribe_topics()
            self._is_connected = True

            async for message in self.client.messages:
                async with self._lock:
                    topic, payload = self.__decode_message(message)
                    await self.on_message(topic, payload)

    async def subscribe(self, topic: str) -> None:
        await self.client.subscribe(topic)
        print(f"[MQTT] Subscribed to {topic}")

    async def subscribe_topics(self) -> None:
        pass

    async def on_message(self, topic: str, payload: Union[str, int]) -> None:
        pass

    async def publish(self, topic: str, payload: Union[str, int, list, None]=None, qos=0, retain=False) -> None:
        await self.client.publish(topic, payload, qos=qos, retain=retain)

    @staticmethod
    def __decode_message(message: Message) -> tuple[str, Union[str, int]]:
        return message.topic.value, MQTTProvider.__parse_payload(message.payload.decode())

    @staticmethod
    def __parse_payload(payload: str) -> Union[str, int]:
        return int(payload) if payload.isnumeric() else payload

    @property
    def is_connected(self) -> bool:
        return self._is_connected

    async def _send_weights(self, topic, weights):
        if weights is not None and len(weights) > 0:
            await self.publish(topic, json.dumps(weights.tolist()))
        else:
            await self.publish(topic, None)

    @staticmethod
    def _unpack_weights(weights):
        weights_list = json.loads(weights)
        return np.array(weights_list)



import time
import board
import busio
import adafruit_scd30


class AdafruitSCD30:
    def __init__(self):
        i2c = busio.I2C(board.SCL, board.SDA)
        self.scd = adafruit_scd30.SCD30(i2c)
        time.sleep(2)

    def read(self):
        if self.scd.data_available:
            co2 = self.scd.CO2
            temperature = self.scd.temperature
            humidity = self.scd.relative_humidity
            return co2, temperature, humidity
        else:
            return None


class ProjectConstants:
    CSV_FIELDS = ["timestamp", "co2", "temperature", "humidity"]
    DATA_DIRECTORY = "./data"
    MONITORING_DIRECTORY = "./monitoring"
    CMD_TOPIC = "sys/cmd/"
    FEEDBACK_TOPIC = "sys/feedback/"
    DATA_TOPIC = "sys/data/"
    FL_ROUND_TOPIC = "sys/fl_round/"
    FL_ROUND_HANDLE = "sys/fl_round_handle/"
    COLLECTION_INTERVAL = 2  # seconds
    FL_ALPHA = 1e-7
    FL_EPOCHS = 3
    FL_GLOBAL_EPOCHS = 20
    MONITORING_INTERVAL = 1  # seconds
    CLIENT_DEVICES = 1



import time
import threading
import psutil
import csv
import os
import json
from datetime import datetime

from kehe_fl.utils.common.project_constants import ProjectConstants


class MonitoringService:
    def __init__(self, directory=ProjectConstants.MONITORING_DIRECTORY, interval=ProjectConstants.MONITORING_INTERVAL):
        self.directory = directory
        self.interval = interval
        self._running = False
        self._thread = None
        self._cpu = []
        self._mem = []
        self._bytes_sent = []
        self._bytes_recv = []
        self._start_net = None
        self._start_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        os.makedirs(self.directory, exist_ok=True)

        self._base_filename = f"run_at_{self._start_time}"
        self.raw_path = os.path.join(self.directory, f"{self._base_filename}.csv")
        self.summary_path = os.path.join(self.directory, f"{self._base_filename}_summary.json")

    def start(self):
        self._running = True
        self._cpu = []
        self._mem = []
        self._bytes_sent = []
        self._bytes_recv = []
        net = psutil.net_io_counters()
        self._start_net = (net.bytes_sent, net.bytes_recv)
        self._thread = threading.Thread(target=self._monitor)
        self._thread.daemon = True
        self._thread.start()

        with open(self.raw_path, 'w', newline='') as f:
            writer = csv.writer(f)
            writer.writerow(['timestamp', 'cpu_percent', 'mem_mb', 'bytes_sent', 'bytes_recv'])

    def _monitor(self):
        while self._running:
            timestamp = time.strftime("%Y-%m-%d %H:%M:%S", time.localtime())
            cpu = psutil.cpu_percent()
            mem = psutil.virtual_memory().used / (1024 * 1024) # Convert to MB
            net = psutil.net_io_counters()
            bytes_sent = net.bytes_sent
            bytes_recv = net.bytes_recv

            self._cpu.append(cpu)
            self._mem.append(mem)
            self._bytes_sent.append(bytes_sent)
            self._bytes_recv.append(bytes_recv)

            with open(self.raw_path, 'a', newline='') as f:
                writer = csv.writer(f)
                writer.writerow([timestamp, cpu, mem, bytes_sent, bytes_recv])
            time.sleep(self.interval)

    def stop(self):
        self._running = False
        if self._thread:
            self._thread.join(timeout=2)

        net = psutil.net_io_counters()
        net_sent_total = net.bytes_sent - self._start_net[0] if self._start_net else 0
        net_recv_total = net.bytes_recv - self._start_net[1] if self._start_net else 0

        summary = {
            "cpu_avg": sum(self._cpu) / len(self._cpu) if self._cpu else 0,
            "cpu_max": max(self._cpu) if self._cpu else 0,
            "cpu_min": min(self._cpu) if self._cpu else 0,
            "mem_avg": sum(self._mem) / len(self._mem) if self._mem else 0,
            "mem_max": max(self._mem) if self._mem else 0,
            "mem_min": min(self._mem) if self._mem else 0,
            "net_sent_total": net_sent_total,
            "net_recv_total": net_recv_total,
            "duration_s": len(self._cpu) * self.interval
        }

        with open(self.summary_path, 'w') as f:
            json.dump(summary, f, indent=2)

        print(f"[MonitoringService] Monitoring stopped. Summary: {summary}")
        print(f"Raw data saved to {self.raw_path}")
        print(f"Summary saved to {self.summary_path}")
        return summary



import asyncio

from kehe_fl.comms.enum.mqtt_cmd_enum import MQTTCmdEnum
from kehe_fl.comms.mqtt_provider import MQTTProvider
from kehe_fl.comms.enum.mqtt_status_enum import MQTTStatusEnum
from kehe_fl.utils.common.project_constants import ProjectConstants
from kehe_fl.utils.service.model_service import ModelService
from kehe_fl.utils.service.monitoring_service import MonitoringService


class MQTTAggServer(MQTTProvider):
    LISTEN_TOPIC = f"{ProjectConstants.FEEDBACK_TOPIC}+"
    FL_ROUND_FEEDBACK_TOPIC = f"{ProjectConstants.FL_ROUND_TOPIC}+"
    clientIds = set()
    commandClientIds = set()
    lastCommand = None
    working = False
    messageQueue = []
    deviceErrorOccurred = False
    weightsArray = []
    newWeights = None

    def __init__(self, broker, port=1883, username=None, password=None):
        super().__init__(broker, port, username, password)
        self.topics = [self.LISTEN_TOPIC, self.FL_ROUND_FEEDBACK_TOPIC]
        self.modelService = ModelService(agg=True)
        self._monitoringService = None
        self._monitoringTask = None
        self._in_queue = asyncio.Queue()
        asyncio.create_task(self._worker())

    async def subscribe_topics(self):
        for topic in self.topics:
            await self.subscribe(topic)
            print(f"[MQTTAggServer] Subscribed to {topic}")

    async def _worker(self):
        while True:
            deviceId, payload = await self._in_queue.get()
            try:
                await self.__handle_data(deviceId, payload)
            finally:
                self._in_queue.task_done()
                self.commandClientIds.add(deviceId)

            self.__after_queue_job_done()

    async def on_message(self, topic: str, payload: str):
        if topic.startswith(ProjectConstants.FL_ROUND_TOPIC[:-1]):
            deviceId = self.__get_device_id_from_topic(topic)
            if deviceId not in self.clientIds:
                return
            if not (payload == "" or payload is None):
                self.weightsArray.append(MQTTProvider._unpack_weights(payload))
                print(f"[MQTTAggServer] Received FL round update from {deviceId}")
            else:
                self.deviceErrorOccurred = True
                print(f"[MQTTAggServer] Received empty FL round update from {deviceId}")
        elif topic.startswith(ProjectConstants.FEEDBACK_TOPIC[:-1]):
            deviceId = MQTTAggServer.__get_device_id_from_topic(topic)
            await self._in_queue.put((deviceId, payload))
        else:
            print(f"[MQTTAggServer] Received unknown topic {topic}: {payload}")

    async def send_update(self, update):
        topic = "sys/update"
        print(f"[MQTTAggServer] Sending update to {topic}: {update}")
        await self.publish(topic, update)

    async def send_command(self, command):
        numCommand = int(command)
        if not any(numCommand == cmd.value for cmd in MQTTCmdEnum):
            print(f"[MQTTAggServer] Invalid command: {command}")
            return

        if numCommand != MQTTCmdEnum.REGISTER_DEVICE.value and len(self.clientIds) == 0:
            print(f"[MQTTAggServer] No devices registered. Cannot send command: {command}")
            return

        self.lastCommand = numCommand
        self.working = True
        if self.lastCommand == MQTTCmdEnum.FL_ROUND.value:
            self._monitoringService = MonitoringService()
            self._monitoringTask = asyncio.create_task(asyncio.to_thread(self._monitoringService.start))
            await self.start_federated_training()
            self._monitoringService.stop()
            await self._monitoringTask
            self._monitoringService = None
            self._monitoringTask = None
            self.working = False
            return

        print(f"[MQTTAggServer] Sending command to {ProjectConstants.CMD_TOPIC}: {command}")
        await self.publish(ProjectConstants.CMD_TOPIC, command)

    async def __handle_data(self, deviceId, data):
        if deviceId not in self.clientIds and self.lastCommand != MQTTCmdEnum.REGISTER_DEVICE.value:
            print(f"[MQTTAggServer] Device {deviceId} not registered.")
            self.deviceErrorOccurred = True
        elif data == MQTTStatusEnum.UNKNOWN_CMD.value:
            MQTTAggServer.__printStatus(deviceId, data)
            self.deviceErrorOccurred = True
        elif self.lastCommand == MQTTCmdEnum.REGISTER_DEVICE.value:
            if data == MQTTStatusEnum.REGISTRATION_SUCCESSFUL.value:
                self.__handle_register(deviceId)
                MQTTAggServer.__printStatus(deviceId, data)
            else:
                MQTTAggServer.__printStatus(deviceId, data)
        elif self.lastCommand == MQTTCmdEnum.START_DATA_COLLECTION.value:
            MQTTAggServer.__printStatus(deviceId, data)
        elif self.lastCommand == MQTTCmdEnum.CHECK_DATA_COUNT.value:
            if data.isnumeric():
                print(f"[MQTTAggServer] Device {deviceId}: Data count {data}")
            else:
                MQTTAggServer.__printStatus(deviceId, data)
        elif self.lastCommand == MQTTCmdEnum.STOP_DATA_COLLECTION.value:
            MQTTAggServer.__printStatus(deviceId, data)
        elif self.lastCommand == MQTTCmdEnum.START_TRAINING.value:
            MQTTAggServer.__printStatus(deviceId, data)
        elif self.lastCommand == MQTTCmdEnum.CHECK_TRAINING_STATUS.value:
            iterationCount, weights = ModelService.unpack_training_status(data)
            if iterationCount is not None and weights is not None:
                print(
                    f"[MQTTAggServer] Device {deviceId}: Training status - Iteration count: {iterationCount}, Weights: {weights}")
            else:
                MQTTAggServer.__printStatus(deviceId, data)
        elif self.lastCommand == MQTTCmdEnum.SEND_UPDATE.value:
            if data.isnumeric():
                print(f"[MQTTAggServer] Device {deviceId}: Update count {data}")
                self.weightsArray.append(data)
            else:
                MQTTAggServer.__printStatus(deviceId, data)

    def __handle_register(self, deviceId):
        if deviceId not in self.clientIds:
            self.clientIds.add(deviceId)
        return

    def __after_queue_job_done(self):
        if self.deviceErrorOccurred:
            self.__handle_clear_worker()
        elif self.lastCommand == MQTTCmdEnum.REGISTER_DEVICE.value and ProjectConstants.CLIENT_DEVICES == len(
                self.clientIds):
            print(f"[MQTTAggServer] All devices registered: {self.clientIds}")
            self.__handle_clear_worker()
        elif self.lastCommand == MQTTCmdEnum.SEND_UPDATE.value and len(
                self.weightsArray) == ProjectConstants.CLIENT_DEVICES:
            print(f"[MQTTAggServer] All devices have sent updates: {self.weightsArray}")
            self.__handle_model_aggregation()
            self.__handle_clear_worker()
        elif self.lastCommand != MQTTCmdEnum.REGISTER_DEVICE.value and len(self.commandClientIds) == len(
                self.clientIds):
            print(f"[MQTTAggServer] All devices have responded to command {self.lastCommand}")
            self.__handle_clear_worker()

    def __handle_clear_worker(self):
        self.deviceErrorOccurred = False
        self.commandClientIds.clear()
        self.lastCommand = None
        self.working = False

    def __handle_model_aggregation(self):
        self.newWeights = self.modelService.aggregate_weights(weights=self.weightsArray)
        self.weightsArray.clear()
        self.modelService.set_weights(self.newWeights)

    async def start_federated_training(self):
        for round_num in range(ProjectConstants.FL_GLOBAL_EPOCHS):
            print(f"[MQTTAggServer] Starting round {round_num + 1}/{ProjectConstants.FL_GLOBAL_EPOCHS}")

            await self.send_fl_round(self.newWeights)

            self.weightsArray.clear()
            self.commandClientIds.clear()

            while len(self.weightsArray) < ProjectConstants.CLIENT_DEVICES:
                if self.deviceErrorOccurred:
                    self.deviceErrorOccurred = False
                    return
                await asyncio.sleep(0.5)  # polling, not elegant but simple

            print(f"[MQTTAggServer] Aggregating weights for round {round_num + 1}")
            self.__handle_model_aggregation()

            if round_num == ProjectConstants.FL_GLOBAL_EPOCHS - 1:
                print("[MQTTAggServer] Finalizing FL round, sending aggregated weights to clients.")
                await self.send_fl_round(self.modelService.get_weights())
            else:
                print(f"[MQTTAggServer] Round {round_num + 1} complete, waiting for next round.")
                self.modelService.set_weights(None)

        print("[MQTTAggServer] Federated training complete.")

    async def send_fl_round(self, weights=None):
        topic = ProjectConstants.FL_ROUND_HANDLE
        print(f"[MQTTAggServer] Starting FL round, sending weights: {weights}")
        await self._send_weights(topic, weights)

    @staticmethod
    def __get_device_id_from_topic(topic):
        return topic.split("/")[-1]

    @staticmethod
    def __printStatus(deviceId, data):
        print(f"[MQTTAggServer] Device {deviceId}: {MQTTStatusEnum.get_status_message(data)}")



import time
from glob import glob

import numpy as np
import pandas as pd

from kehe_fl.utils.common.project_constants import ProjectConstants


class ModelService:
    def __init__(self, agg = False):
        self.__isTraining = False
        self.__weights = None
        self.__x = None
        self.__y = None
        self.__iterationCount = 0

    def __sgd(self, theta):
        x = self.__x
        y = self.__y
        n = len(x)
        alpha = ProjectConstants.FL_ALPHA
        epochs = ProjectConstants.FL_EPOCHS

        for _ in range(epochs):
            indices = np.arange(n)
            np.random.shuffle(indices)
            for i in indices:
                xi = x[i]
                yi = y[i]
                pred = theta[0] + theta[1] * xi
                error = pred - yi
                grad = np.array([error, error * xi])
                theta -= alpha * grad
                self.__iterationCount += 1
        return theta

    def predict(self, x):
        if self.__weights is None:
            print("[MQTTAggServer] No weights")
            return None
        return self.__weights[0] + self.__weights[1] * x

    def start_training(self, data_path: str):
        if not self.__isTraining:
            print("[ModelService] Starting training...")
            self.__isTraining = True

            data_files = glob(f"{data_path}/*.csv")
            if not data_files:
                print("[ModelService] No data files found for training.")
                self.__isTraining = False
                return
            data = pd.concat([pd.read_csv(file) for file in data_files], ignore_index=True)
            self.__x = data["co2"].values
            self.__y = data["temperature"].values

            if self.__weights is None or np.any(np.isnan(self.__weights)):
                self.__weights = np.zeros(2)

            start_time = time.time()
            self.__weights = self.__sgd(self.__weights)
            end_time = time.time()

            print(f"[ModelService] Training completed in {end_time - start_time:.2f} seconds.")
            print(f"[ModelService] Final weights: {self.__weights}")

            self.__isTraining = False
        else:
            print("[ModelService] Training already in progress.")

    def check_training_status(self):
        if self.__isTraining:
            return self.__iterationCount, self.__weights
        else:
            print("[ModelService] No training in progress.")
            return None

    def get_weights(self):
        return self.__weights

    def set_weights(self, weights):
        self.__weights = weights
        return

    #FedAvg
    @staticmethod
    def aggregate_weights(weights):
        return np.mean(weights, axis=0)

    @staticmethod
    def unpack_training_status(data):
        test = data.split(",")
        if len(test) == 2:
            try:
                iterationCount = int(test[0])
                weights = np.array([float(i) for i in test[1].split()])
                return iterationCount, weights
            except ValueError:
                print("[ModelService] Error unpacking training status data")
        else:
            print("[ModelService] Invalid training status data format")


from enum import Enum


class MQTTStatusEnum(Enum):
    REGISTRATION_SUCCESSFUL = 's:r:s'
    REGISTRATION_NOT_SUCCESSFUL = 's:r:n:s'
    DATA_COLLECTION_ALREADY_RUNNING = 's:d:c:a:r'
    DATA_COLLECTION_STARTED = 's:d:c:s'
    DATA_COLLECTION_IN_PROCESS = 's:d:c:i:p'
    DATA_COLLECTION_NOT_RUNNING = 's:d:c:n:r'
    DATA_COLLECTION_STOPPED = 's:d:c:s:t'
    STARTED_TRAINING = 's:s:t'
    TRAINING_ALREADY_STARTED = 's:t:a:s'
    TRAINING_NOT_RUNNING = 's:t:n:r'
    TRAINING_IN_PROCESS = 's:t:i:p'
    UNKNOWN_CMD = 's:u:c'

    @staticmethod
    def get_status_message(code):
        status_messages = {
            MQTTStatusEnum.REGISTRATION_SUCCESSFUL.value: "Connected successfully.",
            MQTTStatusEnum.DATA_COLLECTION_ALREADY_RUNNING.value: "Data collection is already running.",
            MQTTStatusEnum.DATA_COLLECTION_STARTED.value: "Data collection has started.",
            MQTTStatusEnum.DATA_COLLECTION_IN_PROCESS.value: "Data collection is in process.",
            MQTTStatusEnum.DATA_COLLECTION_NOT_RUNNING.value: "Data collection is not running.",
            MQTTStatusEnum.DATA_COLLECTION_STOPPED.value: "Data collection has stopped.",
            MQTTStatusEnum.REGISTRATION_NOT_SUCCESSFUL.value: "Registration failed.",
            MQTTStatusEnum.STARTED_TRAINING.value: "Training has started.",
            MQTTStatusEnum.TRAINING_ALREADY_STARTED.value: "Training is already in progress.",
            MQTTStatusEnum.TRAINING_NOT_RUNNING.value: "Training is not running.",
            MQTTStatusEnum.TRAINING_IN_PROCESS.value: "Training is in process.",
            MQTTStatusEnum.UNKNOWN_CMD.value: "Unknown command.",
        }
        return status_messages.get(code, "Unknown status code.")


import asyncio

from kehe_fl.comms.enum.mqtt_cmd_enum import MQTTCmdEnum
from kehe_fl.comms.enum.mqtt_status_enum import MQTTStatusEnum
from kehe_fl.comms.mqtt_provider import MQTTProvider
from kehe_fl.utils.common.project_constants import ProjectConstants
from kehe_fl.utils.service.data_collection_service import DataCollectionService
from kehe_fl.utils.service.model_service import ModelService
from kehe_fl.utils.service.monitoring_service import MonitoringService


class MQTTDevice(MQTTProvider):
    def __init__(self, broker, deviceId, port=1883, username=None, password=None):
        super().__init__(broker, port, username, password)
        self.deviceId = deviceId
        self.clientTopic = f"{ProjectConstants.FEEDBACK_TOPIC}{deviceId}"
        self.flRoundClientTopic = f"{ProjectConstants.FL_ROUND_TOPIC}{deviceId}"
        self.serverTopics = [ProjectConstants.CMD_TOPIC, ProjectConstants.DATA_TOPIC, ProjectConstants.FL_ROUND_HANDLE]
        self._dataCollectionTask = None
        self._dataCollectionService = None
        self._monitoringService = None
        self._monitoringTask = None
        self._modelTask = None
        self._modelService = None
        self._roundCounter = 0

    async def subscribe_topics(self):
        for topic in self.serverTopics:
            await self.subscribe(topic)

    async def on_message(self, topic: str, payload: int):
        print(f"[MQTTDevice - {self.deviceId}] Received message: {payload} on topic {topic}")

        if topic not in (ProjectConstants.CMD_TOPIC, ProjectConstants.DATA_TOPIC, ProjectConstants.FL_ROUND_HANDLE):
            print(f"[MQTTDevice - {self.deviceId}] Unknown topic {topic}: {payload}")
            return

        if topic == ProjectConstants.DATA_TOPIC:
            await self.check_for_update(payload)
            return

        if topic == ProjectConstants.FL_ROUND_HANDLE:
            if self._roundCounter == 0:
                self._monitoringService = MonitoringService()
                self._monitoringTask = asyncio.create_task(asyncio.to_thread(self._monitoringService.start))
            self._roundCounter += 1
            await self.handle_fl_round(payload)
            if self._roundCounter == ProjectConstants.FL_GLOBAL_EPOCHS:
                self._monitoringService.stop()
                self._monitoringService = None
                await self._monitoringTask
                self._monitoringTask = None
            return

        await self.handle_cmd(payload)

    async def send_data(self, data):
        await self.publish(self.clientTopic, data)

    async def handle_cmd(self, payload):
        if payload == MQTTCmdEnum.START_DATA_COLLECTION.value:
            await self.start_data_collection()
        elif payload == MQTTCmdEnum.CHECK_DATA_COUNT.value:
            await self.check_data_count()
        elif payload == MQTTCmdEnum.STOP_DATA_COLLECTION.value:
            await self._stop_data_collection(withFeedback=True)
        elif payload == MQTTCmdEnum.START_TRAINING.value:
            await self.start_training()
        elif payload == MQTTCmdEnum.CHECK_TRAINING_STATUS.value:
            await self.check_training_status()
        elif payload == MQTTCmdEnum.SEND_UPDATE.value:
            await self.send_update()
        elif payload == MQTTCmdEnum.REGISTER_DEVICE.value:
            await self.send_data(MQTTStatusEnum.REGISTRATION_SUCCESSFUL.value)
        else:
            print("Command not found")

    async def start_data_collection(self):
        if self.__isTraining():
            print(f"[MQTTDevice - {self.deviceId}] Training in process, cannot start data collection")
            await self.send_data(MQTTStatusEnum.TRAINING_IN_PROCESS.value)
            return

        if not self.__isDataCollecting():
            self._dataCollectionService = DataCollectionService(fields=ProjectConstants.CSV_FIELDS,
                                                                path=ProjectConstants.DATA_DIRECTORY,
                                                                interval=ProjectConstants.COLLECTION_INTERVAL)
            self._dataCollectionTask = asyncio.create_task(asyncio.to_thread(self._dataCollectionService.start))
            await self.send_data(MQTTStatusEnum.DATA_COLLECTION_STARTED.value)
        else:
            await self.send_data(MQTTStatusEnum.DATA_COLLECTION_ALREADY_RUNNING.value)
        return

    async def check_data_count(self):
        if self.__isTraining():
            print(f"[MQTTDevice - {self.deviceId}] Training in process, cannot check data count")
            await self.send_data(MQTTStatusEnum.TRAINING_IN_PROCESS.value)
            return

        if self.__isDataCollecting():
            count = self._dataCollectionService.check_data_count()
            await self.send_data(count)
        else:
            print(f"[MQTTDevice - {self.deviceId}] Data collection not running")
            await self.send_data(MQTTStatusEnum.DATA_COLLECTION_NOT_RUNNING.value)
        return

    async def start_training(self):
        if self.__isDataCollecting():
            print(f"[MQTTDevice - {self.deviceId}] Data collection running, stopping it first")
            await self._stop_data_collection()

        if not self.__isTraining():
            self._modelService = ModelService()
            self._modelTask = asyncio.create_task(asyncio.to_thread(self._modelService.start_training,
                                                                    data_path=ProjectConstants.DATA_DIRECTORY))
            print(f"[MQTTDevice - {self.deviceId}] Training started")
            await self.send_data(MQTTStatusEnum.STARTED_TRAINING.value)
        else:
            print(f"[MQTTDevice - {self.deviceId}] Training already running")
            await self.send_data(MQTTStatusEnum.STARTED_TRAINING.value)

    async def check_training_status(self):
        if self.__isDataCollecting():
            print(f"[MQTTDevice - {self.deviceId}] Data collection running, cannot check training status")
            await self.send_data(MQTTStatusEnum.DATA_COLLECTION_IN_PROCESS.value)
            return

        if self.__isTraining():
            await self.send_data(self._modelService.check_training_status())
        else:
            print(f"[MQTTDevice - {self.deviceId}] Training not running")
            await self.send_data(MQTTStatusEnum.TRAINING_NOT_RUNNING.value)
        return

    async def send_update(self):
        if self.__isTraining():
            print(f"[MQTTDevice - {self.deviceId}] Training in process, cannot send update")
            await self.send_data(MQTTStatusEnum.TRAINING_IN_PROCESS.value)
            return

        if self.__isDataCollecting():
            print(f"[MQTTDevice - {self.deviceId}] Data collection in process, cannot send update")
            await self.send_data(MQTTStatusEnum.DATA_COLLECTION_IN_PROCESS.value)
            return

        data = self._modelService.get_weights()
        await self.send_data(data)
        return

    async def check_for_update(self, update):
        if self.__isTraining():
            print(f"[MQTTDevice - {self.deviceId}] Training in process, cannot check for update")
            await self.send_data(MQTTStatusEnum.TRAINING_IN_PROCESS.value)
            return

        if self.__isDataCollecting():
            print(f"[MQTTDevice - {self.deviceId}] Data collection running, cannot check for update")
            await self.send_data(MQTTStatusEnum.DATA_COLLECTION_IN_PROCESS.value)

        if update:
            self._modelService.set_weights(update)
        else:
            print(f"[MQTTDevice - {self.deviceId}] No updates available")
            await self.send_data("No updates available")
        return

    def __isTraining(self):
        if self._modelTask and not self._modelTask.done():
            return True
        return False

    def __isDataCollecting(self):
        if self._dataCollectionTask and not self._dataCollectionTask.done():
            return True
        return False

    async def _stop_data_collection(self, withFeedback=False):
        if self.__isDataCollecting():
            self._dataCollectionService.stop()
            await self._dataCollectionTask
            self._dataCollectionService = None
            print(f"[MQTTDevice - {self.deviceId}] Data collection stopped")
            if withFeedback:
                await self.send_data(MQTTStatusEnum.DATA_COLLECTION_STOPPED.value)
        else:
            print(f"[MQTTDevice - {self.deviceId}] Data collection not running")
            if withFeedback:
                await self.send_data(MQTTStatusEnum.DATA_COLLECTION_NOT_RUNNING.value)
        return

    async def handle_fl_round(self, _weights):
        if self.__isTraining():
            print(f"[MQTTDevice - {self.deviceId}] Training in process, cannot handle FL round")
            return

        if self.__isDataCollecting():
            print(f"[MQTTDevice - {self.deviceId}] Data collection running, cannot handle FL round")
            return

        if not self._modelService:
            self._modelService = ModelService()

        if _weights:
            self._modelService.set_weights(MQTTProvider._unpack_weights(_weights))

        if self._roundCounter <= ProjectConstants.FL_GLOBAL_EPOCHS:
            self._modelService.start_training(data_path=ProjectConstants.DATA_DIRECTORY)
            weights = self._modelService.get_weights()
            await self._send_weights(self.flRoundClientTopic, weights)



from enum import Enum


class MQTTCmdEnum(Enum):
    START_DATA_COLLECTION = 0
    CHECK_DATA_COUNT = 1
    START_TRAINING = 2
    CHECK_TRAINING_STATUS = 3
    SEND_UPDATE = 4
    REGISTER_DEVICE = 5
    STOP_DATA_COLLECTION = 6
    FL_ROUND = 7

    @staticmethod
    def get_command_message(code):
        command_messages = {
            MQTTCmdEnum.START_DATA_COLLECTION.value: "Advised edge devices to start data collection.",
            MQTTCmdEnum.CHECK_DATA_COUNT.value: "Requested edge devices to share data count.",
            MQTTCmdEnum.START_TRAINING.value: "Advised edge devices to start training.",
            MQTTCmdEnum.CHECK_TRAINING_STATUS.value: "Requested edge devices to share training status.",
            MQTTCmdEnum.SEND_UPDATE.value: "Requested edge devices to send update.",
            MQTTCmdEnum.REGISTER_DEVICE.value: "Advised edge devices to register.",
            MQTTCmdEnum.STOP_DATA_COLLECTION.value: "Advised edge devices to stop data collection.",
            MQTTCmdEnum.FL_ROUND.value: "Advised edge devices to start federated learning round.",
        }
        return command_messages.get(code, "Unknown command code.")
